@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2005-2018, University of Amsterdam
+    Copyright (c)  2005-2023, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -42,12 +43,13 @@
             predicate_name/2,           % +Head, -Name
             clause_name/2               % +ClauseRef, -Name
           ]).
-:- use_module(library(lists), [append/3]).
-:- use_module(library(occurs), [sub_term/2]).
-:- use_module(library(debug)).
-:- use_module(library(option)).
-:- use_module(library(listing)).
-:- use_module(library(prolog_source)).
+:- use_module(library(debug),[debugging/1,debug/3]).
+:- autoload(library(listing),[portray_clause/1]).
+:- autoload(library(lists),[append/3]).
+:- autoload(library(occurs),[sub_term/2]).
+:- autoload(library(option),[option/3]).
+:- autoload(library(prolog_source),[read_source_term_at_location/3]).
+
 
 :- public                               % called from library(trace/clause)
     unify_term/2,
@@ -161,6 +163,9 @@ unify_term(X1, X2) :-
 unify_term(X, Y) :-
     float(X), float(Y),
     !.
+unify_term(X, '$BLOB'(_)) :-
+    blob(X, _),
+    \+ atom(X).
 unify_term(X, Y) :-
     string(X),
     is_list(Y),
@@ -232,7 +237,7 @@ try_open_source(File, In) :-
     open_source(File, In),
     !.
 try_open_source(File, In) :-
-    open(File, read, In).
+    open(File, read, In, [reposition(true)]).
 
 
 %!  make_varnames(+ReadClause, +DecompiledClause,
@@ -295,17 +300,38 @@ find_varname(Var, [_|T], Name) :-
 %   This predicate calls the multifile predicate unify_clause_hook/5
 %   with the same arguments to support user extensions.
 %
+%   @arg Module is the source module that   was active when loading this
+%   clause,  which  is  the  same  as  prolog_load_context/2  using  the
+%   `module` context. If this cannot be established  it is the module to
+%   which the clause itself is associated.   The argument may be used to
+%   determine whether or not a specific user transformation is in scope.
+%   See also term_expansion/2,4 and goal_expansion/2,4.
+%
 %   @tbd    This really must be  more   flexible,  dealing with much
 %           more complex source-translations,  falling   back  to  a
 %           heuristic method locating as much as possible.
 
-unify_clause(Read, Read, _, TermPos, TermPos) :-
-    acyclic_term(Read),
-    !.
-                                        % XPCE send-methods
+unify_clause(Read, _, _, _, _) :-
+    var(Read),
+    !,
+    fail.
+unify_clause((RHead :- RBody), (CHead :- CBody), Module, TermPos1, TermPos) :-
+    '$expand':f2_pos(TermPos1, HPos, BPos1,
+                     TermPos2, HPos, BPos2),
+    inlined_unification(RBody, CBody, RBody1, CBody1, RHead,
+                        BPos1, BPos2),
+    RBody1 \== RBody,
+    !,
+    unify_clause2((RHead :- RBody1), (CHead :- CBody1), Module,
+                  TermPos2, TermPos).
+unify_clause(Read, Decompiled, _, TermPos, TermPos) :-
+    Read =@= Decompiled,
+    !,
+    Read = Decompiled.
 unify_clause(Read, Decompiled, Module, TermPos0, TermPos) :-
     unify_clause_hook(Read, Decompiled, Module, TermPos0, TermPos),
     !.
+                                        % XPCE send-methods
 unify_clause(:->(Head, Body), (PlHead :- PlBody), M, TermPos0, TermPos) :-
     !,
     pce_method_clause(Head, Body, PlHead, PlBody, M, TermPos0, TermPos).
@@ -314,19 +340,17 @@ unify_clause(:<-(Head, Body), (PlHead :- PlBody), M, TermPos0, TermPos) :-
     !,
     pce_method_clause(Head, Body, PlHead, PlBody, M, TermPos0, TermPos).
                                         % Unit test clauses
-unify_clause((TH :- Body),
-             (_:'unit body'(_, _) :- !, Body), _,
-             TP0, TP) :-
-    (   TH = test(_,_)
-    ;   TH = test(_)
-    ),
+unify_clause((TH :- RBody), (CH :- !, CBody), Module, TP0, TP) :-
+    plunit_source_head(TH),
+    plunit_compiled_head(CH),
     !,
-    TP0 = term_position(F,T,FF,FT,[HP,BP]),
+    TP0 = term_position(F,T,FF,FT,[HP,BP0]),
+    ubody(RBody, CBody, Module, BP0, BP),
     TP  = term_position(F,T,FF,FT,[HP,term_position(0,0,0,0,[FF-FT,BP])]).
                                         % module:head :- body
 unify_clause((Head :- Read),
              (Head :- _M:Compiled), Module, TermPos0, TermPos) :-
-    unify_clause((Head :- Read), (Head :- Compiled), Module, TermPos0, TermPos1),
+    unify_clause2((Head :- Read), (Head :- Compiled), Module, TermPos0, TermPos1),
     TermPos1 = term_position(TA,TZ,FA,FZ,[PH,PB]),
     TermPos  = term_position(TA,TZ,FA,FZ,
                              [ PH,
@@ -348,12 +372,62 @@ unify_clause(Read, Compiled1, Module, TermPos0, TermPos) :-
     !,
     TermPos2 = term_position(F,T,FF,FT,[ HP, BP ]),
     match_module(Compiled2, Compiled1, Module, TermPos2, TermPos).
-                                        % general term-expansion
-unify_clause(Read, Compiled1, Module, TermPos0, TermPos) :-
+                                               % SSU rules
+unify_clause((Head,RCond => Body), (CHead :- CCondAndBody), Module,
+             term_position(F,T,FF,FT,
+                           [ term_position(_,_,_,_,[HP,CP]),
+                             BP
+                           ]),
+             TermPos) :-
+    split_on_cut(CCondAndBody, CCond, CBody0),
+    !,
+    inlined_unification(RCond, CCond, RCond1, CCond1, Head, CP, CP1),
+    TermPos1 = term_position(F,T,FF,FT, [HP, BP1]),
+    BP2 = term_position(_,_,_,_, [FF-FT, BP]), % Represent (!, Body), placing
+    (   CCond1 == true                         % ! at =>
+    ->  BP1 = BP2,                             % Whole guard is inlined
+        unify_clause2((Head :- !, Body), (CHead :- !, CBody0),
+                      Module, TermPos1, TermPos)
+    ;   mkconj_pos(RCond1, CP1, (!,Body), BP2, RBody, BP1),
+        mkconj_npos(CCond1, (!,CBody0), CBody),
+        unify_clause2((Head :- RBody), (CHead :- CBody),
+                      Module, TermPos1, TermPos)
+    ).
+unify_clause((Head => Body), Compiled1, Module, TermPos0, TermPos) :-
+    !,
+    unify_clause2(Head :- Body, Compiled1, Module, TermPos0, TermPos).
+unify_clause(Read, Decompiled, Module, TermPos0, TermPos) :-
+    unify_clause2(Read, Decompiled, Module, TermPos0, TermPos).
+
+% mkconj, but also unify position info
+mkconj_pos((A,B), term_position(F,T,FF,FT,[PA,PB]), Ex, ExPos, Code, Pos) =>
+    Code = (A,B1),
+    Pos = term_position(F,T,FF,FT,[PA,PB1]),
+    mkconj_pos(B, PB, Ex, ExPos, B1, PB1).
+mkconj_pos(Last, LastPos, Ex, ExPos, Code, Pos) =>
+    Code = (Last,Ex),
+    Pos = term_position(_,_,_,_,[LastPos,ExPos]).
+
+% similar to mkconj, but we should __not__ optimize `true` away.
+mkconj_npos((A,B), Ex, Code) =>
+    Code = (A,B1),
+    mkconj_npos(B, Ex, B1).
+mkconj_npos(A, Ex, Code) =>
+    Code = (A,Ex).
+
+%!  unify_clause2(+Read, +Decompiled, +Module, +TermPosIn, -TermPosOut)
+%
+%   Stratified version to be used after the first match
+
+unify_clause2(Read, Decompiled, _, TermPos, TermPos) :-
+    Read =@= Decompiled,
+    !,
+    Read = Decompiled.
+unify_clause2(Read, Compiled1, Module, TermPos0, TermPos) :-
     ci_expand(Read, Compiled2, Module, TermPos0, TermPos1),
     match_module(Compiled2, Compiled1, Module, TermPos1, TermPos).
                                         % I don't know ...
-unify_clause(_, _, _, _, _) :-
+unify_clause2(_, _, _, _, _) :-
     debug(clause_info, 'Could not unify clause', []),
     fail.
 
@@ -361,6 +435,75 @@ unify_clause_head(H1, H2) :-
     strip_module(H1, _, H),
     strip_module(H2, _, H).
 
+plunit_source_head(test(_,_)) => true.
+plunit_source_head(test(_)) => true.
+plunit_source_head(_) => fail.
+
+plunit_compiled_head(_:'unit body'(_, _)) => true.
+plunit_compiled_head('unit body'(_, _)) => true.
+plunit_compiled_head(_) => fail.
+
+%!  inlined_unification(+BodyRead, +BodyCompiled,
+%!                      -BodyReadOut, -BodyCompiledOut,
+%!                      +HeadRead,
+%!                      +BodyPosIn, -BodyPosOut) is det.
+
+inlined_unification((V=T,RBody0), (CV=CT,CBody0),
+                    RBody, CBody, RHead, BPos1, BPos),
+    inlineable_head_var(RHead, V2),
+    V == V2,
+    (V=T) =@= (CV=CT) =>
+    argpos(2, BPos1, BPos2),
+    inlined_unification(RBody0, CBody0, RBody, CBody, RHead, BPos2, BPos).
+inlined_unification((V=T), (CV=CT),
+                    RBody, CBody, RHead, BPos1, BPos),
+    inlineable_head_var(RHead, V2),
+    V == V2,
+    (V=T) =@= (CV=CT) =>
+    RBody = true,
+    CBody = true,
+    argpos(2, BPos1, BPos).
+inlined_unification((V=T,RBody0), CBody0,
+                    RBody, CBody, RHead, BPos1, BPos),
+    inlineable_head_var(RHead, V2),
+    V == V2,
+    \+ (CBody0 = (G1,_), G1 =@= (V=T)) =>
+    argpos(2, BPos1, BPos2),
+    inlined_unification(RBody0, CBody0, RBody, CBody, RHead, BPos2, BPos).
+inlined_unification((V=_), true,
+                    RBody, CBody, RHead, BPos1, BPos),
+    inlineable_head_var(RHead, V2),
+    V == V2 =>
+    RBody = true,
+    CBody = true,
+    argpos(2, BPos1, BPos).
+inlined_unification(RBody0, CBody0, RBody, CBody, _RHead,
+                    BPos0, BPos) =>
+    RBody = RBody0,
+    BPos  = BPos0,
+    CBody = CBody0.
+
+%!  inlineable_head_var(+Head, -Var) is nondet
+%
+%   True when Var is a variable in  Head   that  may  be used for inline
+%   unification. Currently we only inline direct arguments to the head.
+
+inlineable_head_var(Head, Var) :-
+    compound(Head),
+    arg(_, Head, Var).
+
+split_on_cut((Cond0,!,Body0), Cond, Body) =>
+    Cond = Cond0,
+    Body = Body0.
+split_on_cut((!,Body0), Cond, Body) =>
+    Cond = true,
+    Body = Body0.
+split_on_cut((A,B), Cond, Body) =>
+    Cond = (A,Cond1),
+    split_on_cut(B, Cond1, Body).
+split_on_cut(_, _, _) =>
+    fail.
+
 ci_expand(Read, Compiled, Module, TermPos0, TermPos) :-
     catch(setup_call_cleanup(
               ( set_xref_flag(OldXRef),
@@ -371,7 +514,10 @@ ci_expand(Read, Compiled, Module, TermPos0, TermPos) :-
                 set_prolog_flag(xref, OldXRef)
               )),
           E,
-          expand_failed(E, Read)).
+          expand_failed(E, Read)),
+    compound(TermPos),                  % make sure somthing is filled.
+    arg(1, TermPos, A1), nonvar(A1),
+    arg(2, TermPos, A2), nonvar(A2).
 
 set_xref_flag(Value) :-
     current_prolog_flag(xref, Value),
@@ -448,10 +594,10 @@ a --> { x, y, z }.
 
 %!  ubody(+Read, +Decompiled, +Module, +TermPosRead, -TermPosForDecompiled)
 %
-%   @param Read             Clause read _after_ expand_term/2
-%   @param Decompiled       Decompiled clause
-%   @param Module           Load module
-%   @param TermPosRead      Sub-term positions of source
+%   @arg Read             Clause read _after_ expand_term/2
+%   @arg Decompiled       Decompiled clause
+%   @arg Module           Load module
+%   @arg TermPosRead      Sub-term positions of source
 
 ubody(B, DB, _, P, P) :-
     var(P),                        % TBD: Create compatible pos term?
@@ -464,15 +610,40 @@ ubody(B, C, _, P, P) :-
 ubody(X0, X, M, parentheses_term_position(_, _, P0), P) :-
     !,
     ubody(X0, X, M, P0, P).
-ubody(X, call(X), _,                    % X = call(X)
+ubody(X, Y, _,                    % X = call(X)
       Pos,
       term_position(From, To, From, To, [Pos])) :-
+    nonvar(Y),
+    Y = call(X),
     !,
     arg(1, Pos, From),
     arg(2, Pos, To).
+ubody(A, B, _, P1, P2) :-
+    nonvar(A), A = (_=_),
+    nonvar(B), B = (LB=RB),
+    A =@= (RB=LB),
+    !,
+    P1 = term_position(F,T, FF,FT, [PL,PR]),
+    P2 = term_position(F,T, FF,FT, [PR,PL]).
+ubody(A, B, _, P1, P2) :-
+    nonvar(A), A = (_==_),
+    nonvar(B), B = (LB==RB),
+    A =@= (RB==LB),
+    !,
+    P1 = term_position(F,T, FF,FT, [PL,PR]),
+    P2 = term_position(F,T, FF,FT, [PR,PL]).
 ubody(B, D, _, term_position(_,_,_,_,[_,RP]), TPOut) :-
     nonvar(B), B = M:R,
     ubody(R, D, M, RP, TPOut).
+ubody(B, D, M, term_position(_,_,_,_,[RP0,RP1]), TPOut) :-
+    nonvar(B), B = (B0,B1),
+    (   maybe_optimized(B0),
+        ubody(B1, D, M, RP1, TPOut)
+    ->  true
+    ;   maybe_optimized(B1),
+        ubody(B0, D, M, RP0, TPOut)
+    ),
+    !.
 ubody(B0, B, M,
       brace_term_position(F,T,A0),
       Pos) :-
@@ -501,6 +672,8 @@ ubody(Read, Decompiled, Module, TermPosRead, TermPosDecompiled) :-
 ubody(X0, X, M,
       term_position(F,T,FF,TT,PA0),
       term_position(F,T,FF,TT,PA)) :-
+    callable(X0),
+    callable(X),
     meta(M, X0, S),
     !,
     X0 =.. [_|A0],
@@ -510,7 +683,9 @@ ubody(X0, X, M,
 ubody(X0, X, M,
       term_position(F,T,FF,TT,PA0),
       term_position(F,T,FF,TT,PA)) :-
-    expand_goal(X0, X, M, PA0, PA).
+    expand_goal(X0, X1, M, PA0, PA),
+    X1 =@= X,
+    X1 = X.
 
                                         % 5.7.X optimizations
 ubody(_=_, true, _,                     % singleton = Any
@@ -548,6 +723,11 @@ ubody_elem(0, G0, G, M, PA0, PA) :-
     ubody(G0, G, M, PA0, PA).
 ubody_elem(_, G, G, _, PA, PA).
 
+%!  conj(+GoalTerm, +PositionTerm, -GoalList, -PositionList)
+%
+%   Turn a conjunctive body into a list   of  goals and their positions,
+%   i.e., removing the positions of the (,)/2 terms.
+
 conj(Goal, Pos, GoalList, PosList) :-
     conj(Goal, Pos, GoalList, [], PosList, []).
 
@@ -571,6 +751,8 @@ conj((!,(S=SR)), F-T, [!,S=SR|TG], TG, [F-T,F1-T1|TP], TP) :-
 conj(A, P, [A|TG], TG, [P|TP], TP).
 
 
+%!  mkconj(+Decompiled, +Module, -Position, +ReadGoals, +ReadPositions)
+
 mkconj(Goal, M, Pos, GoalList, PosList) :-
     mkconj(Goal, M, Pos, GoalList, [], PosList, []).
 
@@ -581,7 +763,25 @@ mkconj(Conj, M, term_position(0,0,0,0,[PA,PB]), GL, TG, PL, TP) :-
     mkconj(A, M, PA, GL, TGA, PL, TPA),
     mkconj(B, M, PB, TGA, TG, TPA, TP).
 mkconj(A0, M, P0, [A|TG], TG, [P|TP], TP) :-
-    ubody(A, A0, M, P, P0).
+    ubody(A, A0, M, P, P0),
+    !.
+mkconj(A0, M, P0, [RG|TG0], TG, [_|TP0], TP) :-
+    maybe_optimized(RG),
+    mkconj(A0, M, P0, TG0, TG, TP0, TP).
+
+maybe_optimized(debug(_,_,_)).
+maybe_optimized(assertion(_)).
+maybe_optimized(true).
+
+%!  argpos(+N, +PositionTerm, -ArgPositionTerm) is det.
+%
+%   Get the position for the nth argument of PositionTerm.
+
+argpos(N, parentheses_term_position(_,_,PosIn), Pos) =>
+    argpos(N, PosIn, Pos).
+argpos(N, term_position(_,_,_,_,ArgPos), Pos) =>
+    nth1(N, ArgPos, Pos).
+argpos(_, _, _) => true.
 
 
                  /*******************************
@@ -719,9 +919,10 @@ the clause as defined in init.pl
 expand_goal(G, call(G), _, P, term_position(0,0,0,0,[P])) :-
     var(G),
     !.
-expand_goal(G, G, _, P, P) :-
+expand_goal(G, G1, _, P, P) :-
     var(G),
-    !.
+    !,
+    G1 = G.
 expand_goal(M0, M, Module, P0, P) :-
     meta(Module, M0, S),
     !,
